/**
 *
 *			MemManager.cpp
 *			Ryan Medeiros
 */
#include "MemManager.h"

#include <iostream>
#include <cassert>

#ifdef new
#undef new
#endif

#ifndef UNREFERENCED_PARAMETER
#define UNREFERENCED_PARAMETER(P)	(P)
#endif

#define MAGICNUMBER					(0xCAFEBABE)
#define MAGICMARKER					(0xFACEBABE)
#define WHIPEMARKER					(0xBAADBAAD)
#define SETUPHEADER										\
	myHeader   = (Header*) _mem;						\
	firstFree  = (MemHeader*) myHeader->firstFree;		\
	firstAlloc = (MemHeader*) myHeader->firstAlloc;

const MemManager* MemManager::_firstOrGlobalManager = 0;

namespace
{
	/** 
	 * The following are helper classes/functions which are
	 * utilized by the memory manager class.  They are defined
	 * here simply to make the header cleaner and hide some
	 * private static functions.
	 */

	struct Header
	{
		void* firstFree;
		void* firstAlloc;

#ifdef _DEBUG
		unsigned int currTime;
		size_t currAllocated;
		size_t highWater;
#endif
	};

	struct MemHeader
	{
		int        magic;
		size_t     size;
		MemHeader* next;

#ifdef _DEBUG
		int          tag;
		unsigned int whenCreated;
		char*        fileName;
		unsigned int line;
#endif _DEBUG
	};

	struct Marker
	{
		int        magic;
		MemHeader* myHeader;
	};

	void WriteMarker(MemHeader *which)
	{
		Marker* marker = (Marker*)((char*)which + sizeof(MemHeader) + which->size - sizeof(Marker));
		marker->magic = MAGICMARKER;
		marker->myHeader = which;
	}

	void WipeMarker(MemHeader *which)
	{
		Marker* marker = (Marker*)((char*)which + sizeof(MemHeader) + which->size - sizeof(Marker));
		marker->magic = WHIPEMARKER;
		marker->myHeader = NULL;
	}

	void MarkFreeMem(MemHeader* toFree)
	{
		for(	int* i = (int*) ( (char*)toFree + sizeof(MemHeader) ) ;
				i < (int*) ( (char*)toFree + sizeof(MemHeader) + toFree->size - sizeof(Marker)) ;
				++i )
			*i = 0xDEADDEAD;
	}

	void Coalesce(void *mem)
	{
		Header     *header = (Header*) mem;
		MemHeader **firstFree = (MemHeader**) &header->firstFree;
		MemHeader **found, *other;
		Marker     *marker;

		for( found = firstFree ; *found ; found = &(*found)->next )
		{
			// Next Block in List Immediately Follows (faster than 3rd case)
			if( *found && (char*)(*found)->next == (char*) (*found) + sizeof(MemHeader) + (*found)->size )
			{
				other = (*found)->next;
				(*found)->size += other->size + sizeof(MemHeader);
				(*found)->next = other->next;
				WriteMarker(*found);
#ifdef _DEBUG
				MarkFreeMem(*found);
#endif
				found = firstFree;
				continue;
			}

			// Immediately preceding block
			marker = (Marker*) ((char*)(*found) - sizeof(Marker));

			if( marker->magic == MAGICMARKER && marker->myHeader->magic == MAGICNUMBER 
				&& (marker->myHeader->size & 1 ^ 1))
			{
				other = marker->myHeader;
				other->size += (*found)->size + sizeof(MemHeader);
				*found = (*found)->next;
				WriteMarker(other);
#ifdef _DEBUG
				MarkFreeMem(other);
#endif
				found = firstFree;
				continue;
			}

			// Immediately following block
			other = (MemHeader*) ((char*)(*found) + (*found)->size + sizeof(MemHeader));
			
			if( other->magic == MAGICNUMBER && (other->size & 1 ^ 1) )
			{
				MemHeader **pointsToOther;

				for( pointsToOther = firstFree ; *pointsToOther != other ; pointsToOther = & (*pointsToOther)->next )
					assert( (*pointsToOther)->magic == MAGICNUMBER ); // Verify Heap in Debug

				*pointsToOther = other->next;
				(*found)->size += other->size + sizeof(MemHeader);
				WriteMarker(*found);
#ifdef _DEBUG
				MarkFreeMem(*found);
#endif
				found = firstFree;
				continue;
			}
		}
	}
}

MemManager::MemManager(void* mem_ptr, size_t size)
 :	_mem(mem_ptr)
{
	assert( mem_ptr != NULL && size > (sizeof(Header)+sizeof(MemHeader)) );

	Header* myHeader = (Header*) _mem;
	MemHeader* firstFree = (MemHeader*) ((char*)_mem + sizeof(Header));

	myHeader->firstFree  = firstFree;
	myHeader->firstAlloc = NULL;

	firstFree->magic = MAGICNUMBER;
	firstFree->size  = size - sizeof(Header) - sizeof(MemHeader) - sizeof(int);
	firstFree->next  = NULL;

	// Writing a unique int to the end of a manager's block.  This prevents
	// two adjoining memory managers from coalescing each other's memory.
	*(int*)( (char*) firstFree + firstFree->size +sizeof(MemHeader)) = 0xBEEFCACE;

#ifdef _DEBUG
	myHeader->currTime = 0;
	myHeader->currAllocated = 0;
	myHeader->highWater = 0;
#endif

	if(!_firstOrGlobalManager)
		_firstOrGlobalManager = this;
}

MemManager::~MemManager()
{
}

void MemManager::SetGlobalManager(const MemManager* which)
{
	assert(which);
	_firstOrGlobalManager = which;
}

void* MemManager::Alloc(size_t size, char* filenameptr /*= 0 */, 
						unsigned int line /*= 0 */, int tag /*= 'BASE'*/) const
{
	assert(this);
	if( size < sizeof(Marker) ) size = sizeof(Marker);
	if( !(size%sizeof(int) )  ) size += size%sizeof(int);

	Header*	   myHeader;
	MemHeader* firstFree;
	MemHeader* firstAlloc;
	MemHeader **found, *created;

	SETUPHEADER;

	for( found = (MemHeader**)& myHeader->firstFree ; 
			*found && (*found)->size < size ; 
			found = &(*found)->next )
		assert( (*found)->magic == MAGICNUMBER ) ;  // Verify Heap

	if( !*found ) throw std::bad_alloc("Out of Memory");
	assert( (*found)->magic == MAGICNUMBER && !((*found)->size & 1) ); // Check for corrupt Heap

	if( size + sizeof(MemHeader) < (*found)->size / 4 )
	{
		// Small allocation, place at end of block
		created = (MemHeader*)( (char*) (*found) + (*found)->size - size);
		(*found)->size -= size + sizeof(MemHeader);
		created->size = size;
		WriteMarker(*found);
	}
	else if( size + sizeof(MemHeader) + sizeof(Marker) <= (*found)->size )
	{
		// Large allocation, place at front of block
		created = (*found);
		*found = (MemHeader*)((char*)created + size + sizeof(MemHeader));
		(*found)->magic = MAGICNUMBER;
		(*found)->size  = created->size - size - sizeof(MemHeader);
		(*found)->next  = created->next;
		created->size   = size;
		WriteMarker(*found);
	}
	else // Don't Split!!
	{
		created = *found;
		*found = (*found)->next;
	}

	created->magic = MAGICNUMBER;
	created->next  = firstAlloc;
	myHeader->firstAlloc = created;
	WipeMarker(created);
	created->size  |= 1;

#ifdef _DEBUG
	created->tag = tag;
	created->whenCreated = myHeader->currTime++;
	created->fileName = filenameptr;
	created->line = line;

	for( int* i = (int*) ( (char*)created + sizeof(MemHeader) ) ;
		 i < (int*) ( (char*)created + sizeof(MemHeader) + (created->size ^ 1)) ;
		 ++i )
		 *i = 0xFEEDFEED;

	myHeader->currAllocated += size;
	myHeader->highWater = myHeader->highWater < myHeader->currAllocated ? myHeader->currAllocated : myHeader->highWater;
#else
	UNREFERENCED_PARAMETER(filenameptr);
	UNREFERENCED_PARAMETER(line);
	UNREFERENCED_PARAMETER(tag);
#endif

	return (char*)created + sizeof(MemHeader);
}

void MemManager::Free(void* ptr) const throw() 
{
	assert(this);
	Header*	   myHeader;
	MemHeader* firstFree;
	MemHeader* firstAlloc;
	MemHeader **found, *toFree;

	SETUPHEADER;

	toFree = (MemHeader*) ((char*)ptr - sizeof(MemHeader));
	assert(toFree->magic == MAGICNUMBER);   // Not a Block
	assert(toFree->size & 1);				// Double Free
	toFree->size ^= 1;

	for( found = (MemHeader**)&myHeader->firstAlloc ; 
		 *found && (*found) != toFree ; 
		 found = &(*found)->next )
		assert( (*found)->magic == MAGICNUMBER ) ;

	assert(*found); // Bad Heap
	(*found) = toFree->next;

	toFree->next = firstFree;
	myHeader->firstFree = toFree;
	WriteMarker(toFree);

#ifdef _DEBUG
	myHeader->currAllocated -= toFree->size;
	MarkFreeMem(toFree);
	myHeader->currTime++;
#endif

	Coalesce(_mem);
}

unsigned int MemManager::GetMemTime()
{
#ifdef _DEBUG
	Header *myHeader = (Header*) _mem;
	return myHeader->currTime;
#else
	return 0;
#endif
}

bool MemManager::NoTemporalLeaks(unsigned int since)
{
#ifdef _DEBUG
	Header*	   myHeader;
	MemHeader* firstFree;
	MemHeader* firstAlloc;
	MemHeader* alloced;
	SETUPHEADER;

	for( alloced = firstAlloc ; alloced ; alloced = alloced->next )
	{
		if(alloced->whenCreated >= since)
			return false;
	}
	return true;
#else
	UNREFERENCED_PARAMETER(since);
	return true;
#endif
}

void MemManager::WriteDebug(std::ostream& out)
{
#ifdef _DEBUG
	Header*	   myHeader;
	MemHeader* firstFree;
	MemHeader* firstAlloc;
	MemHeader* alloced;
	SETUPHEADER;

	out << "\tPrinting Debug information for memory manager at " << this << std::endl;
	out << "\tManaging memory block at " << _mem << "\n\n";
	out << "Most Ever Allocated:      " << myHeader->highWater << std::endl;
	out << "Total Number new/deletes: " << myHeader->currTime << "\n\n";

	if( myHeader->firstAlloc )
	{
		out << "Memory Leaks found!\n";
		out << "Total memory allocated: " << myHeader->currAllocated << std::endl;

		for( alloced = firstAlloc ; alloced ; alloced = alloced->next )
		{
			out << "\nBlock found at address: " << (void*)((char*) alloced + sizeof(MemHeader));
			out << "\nNew/Delete number:      " << alloced->whenCreated;
			out << "\nBlock of size:          " << (alloced->size ^1);
			out << "\nTag:                    " << ((char*)&alloced->tag)[3] << ((char*)&alloced->tag)[2]
												<< ((char*)&alloced->tag)[1] << ((char*)&alloced->tag)[0];
			out << "\nFile: " << alloced->fileName << "\nLine: " << alloced->line <<std::endl;
		}

	}
	else
		out << "No Leaks!\n";

#else
	UNREFERENCED_PARAMETER(out);
#endif
}

void* operator new ( size_t size, char* fileName, unsigned int line, 
					const MemManager& myMM /*= *MemManager::_firstOrGlobalManager */,
					int tag /*= 'BASE' */)
{
	return myMM.Alloc(size, fileName, line, tag);
}

void* operator new [] ( size_t size, char* fileName, unsigned int line, 
					   const MemManager& myMM /*= *MemManager::_firstOrGlobalManager */,
					   int tag /*= 'BASE' */)
{
	return myMM.Alloc(size, fileName, line, tag);
}

void operator delete (void* p, char*, unsigned int, const MemManager& myMM, int)
{
	myMM.Free(p);
}

void operator delete [] (void* p, char*, unsigned int, const MemManager& myMM, int)
{
	myMM.Free(p);
}

void operator delete(void* p)
{
	MemManager::_firstOrGlobalManager->Free(p);
}

void operator delete [](void* p)
{
	MemManager::_firstOrGlobalManager->Free(p);
}

void operator delete (void* p, const MemManager& myMM)
{
	myMM.Free(p);
}

void operator delete [] (void* p, const MemManager& myMM)
{
	myMM.Free(p);
}