/**
 *
 *			MemManager.h
 *			Ryan Medeiros
 *
 *	This class handles memory management for a larger application.
 *	Provides debug information and throws exceptions on out-of-
 *	memory issues. Customized for use with a game engine.
 *
 *	Use:
 *		Creation requires both a memory location and the size of
 *	the memory block being managed. Must be contiguous memory.
 *
 */
#pragma once
#ifndef _MEMMANAGER_H_
#define _MEMMANAGER_H_

#include<iostream>

#ifdef _DEBUG
#define new				new(__FILE__,__LINE__)
#define NEW(MM)			new(__FILE__,__LINE__,MM)
#else
#define new				new(NULL,NULL)
#define NEW(MM)			new(NULL,NULL,MM)
#endif

#define DELETE(PTR, MM)	operator delete(PTR,MM)

class MemManager
{
public:
	MemManager(void* mem_ptr, size_t size) /* throw(std::bad_alloc) */;
	~MemManager();

	void *Alloc(size_t size, char* filenameptr = 0, unsigned int line = 0, int tag = 'NONE') const;
	void Free(void* ptr) const throw(); // assert on double free

	static void SetGlobalManager(const MemManager* which);

	// The following three functions only perform in debug mode,
	// in release mode they have no contents (NoTemporalLeaks simply
	// returns true, GetMemTime: 0).
	//
	unsigned int GetMemTime();
	bool NoTemporalLeaks(unsigned int since);
	void WriteDebug(std::ostream& out);

private:

	void*  _mem;

	static const MemManager* _firstOrGlobalManager;

	MemManager();
	MemManager(const MemManager&);
	void operator =(const MemManager&);

// This undef exists so that new can be defined at the top of
// the header, it is redefined at the bottom.
#undef new

	friend void* operator new (size_t, char*, unsigned int, const MemManager&, int);
	friend void* operator new [] (size_t, char*, unsigned int, const MemManager&, int);

	friend void operator delete (void* p);
	friend void operator delete [] (void* p);

};

void* operator new ( size_t size, char* fileName, unsigned int line, 
					   const MemManager& myMM = *MemManager::_firstOrGlobalManager, int tag = 'NONE' );
void* operator new [] ( size_t size, char* fileName, unsigned int line, 
					   const MemManager& myMM = *MemManager::_firstOrGlobalManager, int tag = 'NONE'  );
void operator delete (void* p, char*, unsigned int, const MemManager&, int);
void operator delete [] (void* p, char*, unsigned int, const MemManager&, int);

void operator delete (void* p, const MemManager& myMM);
void operator delete [] (void* p, const MemManager& myMM);

#ifdef _DEBUG
#define new				new(__FILE__,__LINE__)
#else
#define new				new(NULL,NULL)
#endif

#endif
